from decimal import Decimal
from typing import List

from nodekit._internal.models.node_engine.bonus_policy import BonusRule, ConstantBonusRule
from nodekit._internal.models.node_engine.node_graph import NodeResult


# %% BonusPolicy Rule Engine
def compute_bonus(
        bonus_rules: List[BonusRule],
        node_results: List[NodeResult]
) -> Decimal:

    calculated_amount = Decimal('0')

    for node_result in node_results:
        action = node_result.action

        # Perform scan through rules
        for rule in bonus_rules:
            # Dynamic dispatch
            if isinstance(rule, ConstantBonusRule):
                rule: ConstantBonusRule
                if action.sensor_id == rule.bonus_rule_parameters.sensor_id:
                    # Check currency match
                    calculated_amount += Decimal(rule.bonus_rule_parameters.bonus_amount_usd)

    # Clip at minimum of 0:
    if calculated_amount < Decimal('0'):
        calculated_amount = Decimal('0.00')

    return calculated_amount


__all__ = [
    # Reusable structural types used as field values in models:
    'TextContent',
    'Board',
    'BoardRectangle',
    'BoardLocation',
    'Timespan',

    'RuntimeMetrics',
    'PixelArea',
]

from nodekit._internal.models.node_engine.fields import (
    TextContent,
    BoardRectangle,
    BoardLocation,
    Timespan,
)

from nodekit._internal.models.node_engine.runtime_metrics import (
    PixelArea,
    RuntimeMetrics,
)

from nodekit._internal.models.node_engine.board import Board





from typing import Literal, Annotated, Union

import pydantic

from nodekit._internal.models.fields import DatetimeUTC
from nodekit._internal.models.node_engine.base import NullValue
from nodekit._internal.models.node_engine.node_graph import NodeResult

from uuid import UUID


# %%
import enum
class EventTypeEnum(str, enum.Enum):
    StartEvent = 'StartEvent'
    NodeResultEvent = 'NodeResultEvent'
    EndEvent = 'EndEvent'

# %%
class BaseEvent(pydantic.BaseModel):
    event_id: UUID
    run_id: UUID
    event_type: EventTypeEnum
    event_payload: NullValue
    event_timestamp: DatetimeUTC


# %%
class StartEvent(BaseEvent):
    event_type: Literal[EventTypeEnum.StartEvent] = EventTypeEnum.StartEvent


class EndEvent(BaseEvent):
    event_type: Literal[EventTypeEnum.EndEvent] = EventTypeEnum.EndEvent


# %%
class NodeResultEvent(BaseEvent):
    event_type: Literal[EventTypeEnum.NodeResultEvent] = EventTypeEnum.NodeResultEvent
    event_payload: NodeResult

# %%
Event = Annotated[
    Union[
        StartEvent,
        NodeResultEvent,
        EndEvent,
    ],
    pydantic.Field(discriminator='event_type')
]


# %% Expected server communication protocol:
SubmitEventRequest = Event

class SubmitEventResponse(pydantic.BaseModel):
    redirect_url: str | None = pydantic.Field(
        description="The URL to which the Participant browser should redirect after submitting this Event. If None, no redirect is needed.",
    )



from abc import ABC
from typing import Any

import pydantic

from nodekit._internal.models.node_engine.base import DslModel
from nodekit._internal.models.node_engine.fields import Timespan, SensorId, CardId
from uuid import uuid4


class BaseSensor(DslModel, ABC):
    """
    A Sensor represents a listener for Participant behavior. Sensors are bound to a specific Card, or to the Board.
    When a Sensor is triggered, it emits a fully qualified Action.
    """

    # Sensor
    sensor_id: SensorId = pydantic.Field(default_factory=uuid4)
    sensor_type: str
    sensor_parameters: Any
    # Temporal
    sensor_timespan: Timespan

    # Sensor target
    card_id: CardId | None = pydantic.Field(description='Identifier of the Entity (Card or Board) that this Sensor is attached to.')



# This module exposes the public API for the psykit package.

__all__ = [
    'VERSION',

    # Top-level models:
    'NodeGraph',
    'Node',
    'NodeResult',

    # Namespaced models:
    'actions',
    'assets',
    'cards',
    'effects',
    'reinforcer_maps',
    'sensors',
    'bonus_rules',
    'events',

    # Types:
    'types',

    # Runtime API:
    'compile',
]

# Incoming models:
from nodekit._internal.models.node_engine.node_graph import (
    NodeGraph,
    Node,
    NodeResult
)

import nodekit.cards as cards
import nodekit.assets as assets
import nodekit.effects as effects
import nodekit.reinforcer_maps as reinforcer_maps
import nodekit.sensors as sensors
import nodekit.actions as actions
import nodekit.bonus_rules as bonus_rules
import nodekit.types as types
import nodekit.compile as compile
import nodekit.events as events

VERSION = '0.0.1'



from abc import ABC
from typing import Any

from nodekit._internal.models.node_engine.base import DslModel
from nodekit._internal.models.node_engine.fields import SensorId


class BaseReinforcerMap(DslModel, ABC):
    """
    Represents a map from a fully qualified Action emitted by a particular Sensor to an Outcome.
    """
    reinforcer_map_type: str
    reinforcer_map_parameters: Any
    sensor_id: SensorId



from typing import Literal

from nodekit._internal.models.node_engine.base import DslModel, NullParameters
from nodekit._internal.models.node_engine.reinforcer_maps.base import BaseReinforcerMap
from nodekit._internal.models.node_engine.reinforcer_maps.reinforcer.reinforcer import Reinforcer

from typing import Annotated, Union

import pydantic


# %%
class ConstantReinforcerMap(BaseReinforcerMap):
    class Parameters(DslModel):
        reinforcer: Reinforcer = pydantic.Field(description='The Outcome to return for any Action emitted by the Sensor it is attached to.')

    """
    An OutcomeMap which always returns the same Outcome for any Action emitted by the Sensor it is attached to.
    """
    reinforcer_map_type: Literal['ConstantReinforcerMap'] = 'ConstantReinforcerMap'
    reinforcer_map_parameters: Parameters


# %%
class NullReinforcerMap(BaseReinforcerMap):
    """
    A convenience class which represents an ReinforcerMap which yields a NullReinforcer.
    """
    reinforcer_map_type: Literal['NullReinforcerMap'] = 'NullReinforcerMap'
    reinforcer_map_parameters: NullParameters = pydantic.Field(
        default_factory=NullParameters, frozen=True
    )


# %%
ReinforcerMap = Annotated[
    Union[
        NullReinforcerMap,
        ConstantReinforcerMap,
        # Add other OutcomeMap types here as needed
    ],
    pydantic.Field(discriminator='reinforcer_map_type')
]



import pydantic


class DslModel(pydantic.BaseModel):
    pass



class NullParameters(DslModel):
    """
    A sentinel model for *_parameter fields which do not require specification.
    """
    pass


class NullValue(DslModel):
    """
    A sentinel model for *_value fields which do not require specification.
    """
    pass





import datetime
from decimal import Decimal
from typing import Literal, Annotated

import pydantic


# %%
def _ensure_monetary_amount_precision(value: str) -> str:
    SubcentMonetaryAmountAdapter = pydantic.TypeAdapter(
        Annotated[Decimal, pydantic.Field(decimal_places=5)]
    )
    d = SubcentMonetaryAmountAdapter.validate_python(value)
    return str(d)


MonetaryAmountUsd = Annotated[
    str,
    pydantic.Field(description='An arbitrary amount of money in USD, including negative amounts, represented as a string with at most five decimal places, e.g., "1.00001".'),
    pydantic.AfterValidator(_ensure_monetary_amount_precision)
]


# %%
def _ensure_payable_monetary_amount(value: str) -> str:
    PayableMonetaryAmountAdapter = pydantic.TypeAdapter(
        Annotated[Decimal, pydantic.Field(decimal_places=5)]
    )
    d = PayableMonetaryAmountAdapter.validate_python(value)
    return str(d)


PayableMonetaryAmountUsd = Annotated[
    str,
    pydantic.Field(description='A semi-positive amount of money in USD that is payable to a worker, represented as a string with at most two decimal places, e.g., "1.00". This amount must be at least "0.01".'),
    pydantic.AfterValidator(_ensure_payable_monetary_amount)
]


# %% Timestamps
def ensure_utc(t: datetime.datetime) -> datetime.datetime:
    # Ensures that a datetime is timezone-aware and in UTC.
    if t.tzinfo is None:
        raise ValueError(f"Datetime must be timezone-aware: {t}")
    return t.astimezone(datetime.timezone.utc)


DatetimeUTC = Annotated[
    datetime.datetime,
    pydantic.Field(description='A timezone-aware datetime in UTC.'),
    pydantic.AfterValidator(ensure_utc)
]

# %%
SHA256 = Annotated[str, pydantic.Field(pattern=r'^[a-f0-9]{64}$')]

MimeType = Literal[
    'image/png',
    # Add other supported mime types here as needed
]




from typing import List

import pydantic

from nodekit._internal.models.node_engine.base import DslModel
from nodekit._internal.models.node_engine.cards.cards import Card


class Reinforcer(DslModel):

    reinforcer_cards: List[Card] = pydantic.Field(
        description='Cards which configure the (noninteractive) Reinforcer delivery phase.'
    )



__all__ = [
    'FixationPointCard',
    'MarkdownPagesCard',
    'ImageCard',
    'TextCard',
]

from nodekit._internal.models.node_engine.cards.cards import (
    FixationPointCard,
    MarkdownPagesCard,
    ImageCard,
    TextCard,
)



import mimetypes
from abc import ABC
from typing import Self, Annotated, Union, Literal

import pydantic

from nodekit._internal.models.fields import SHA256, MimeType


# %%
class BaseAssetLink(pydantic.BaseModel, ABC):
    mime_type: MimeType
    sha256: SHA256
    asset_url: pydantic.AnyHttpUrl

    @pydantic.model_validator(mode='after')
    def check_url(self) -> Self:
        """
        Validate that the URL ends with the expected file extension
        """
        extension = mimetypes.guess_extension(type=self.mime_type, strict=True)
        if not extension:
            raise ValueError(f"Could not determine file extension for mime type {self.mime_type}.")

        if not str(self.asset_url).endswith(extension):
            raise ValueError(f"AssetLink {self.asset_url} does not end with the expected file extension {extension}.")

        return self


# %%
class ImageLink(BaseAssetLink):
    mime_type: Literal['image/png'] = 'image/png'


# %%
AssetLink = Annotated[
    Union[
        ImageLink,
        # Add other AssetLink types here as needed
    ],
    pydantic.Field(discriminator='mime_type')
]



from abc import ABC
from typing import Any

import pydantic

from nodekit._internal.models.node_engine.base import DslModel
from nodekit._internal.models.node_engine.fields import TimePointMsec, SensorId


class BaseAction(DslModel, ABC):
    sensor_id: SensorId = pydantic.Field(
        description='Identifier of the Sensor that emitted this Action.'
    )
    action_type: str
    action_value: Any
    reaction_time_msec: TimePointMsec = pydantic.Field(
        description='Measured from the onset of the earliest possible time the Action could be emitted.'
    )



from typing import List, Self
from uuid import uuid4

import pydantic

from nodekit._internal.models.fields import (
    PayableMonetaryAmountUsd, DatetimeUTC
)
from nodekit._internal.models.node_engine.board import Board
from nodekit._internal.models.node_engine.bonus_policy import BonusRule
from nodekit._internal.models.node_engine.cards.cards import Card
from nodekit._internal.models.node_engine.effects.base import Effect
from nodekit._internal.models.node_engine.fields import NodeId
from nodekit._internal.models.node_engine.reinforcer_maps.reinforcer_maps import ReinforcerMap
from nodekit._internal.models.node_engine.runtime_metrics import RuntimeMetrics
from nodekit._internal.models.node_engine.sensors.actions.actions import Action
from nodekit._internal.models.node_engine.sensors.sensors import Sensor


# %% Node
class Node(pydantic.BaseModel):
    node_id: NodeId = pydantic.Field(default_factory=uuid4)

    board: Board = pydantic.Field(
        default_factory=Board
    )

    cards: List[Card] = pydantic.Field(
        description="List of Cards that will be placed on the Board, in back-to-front order (i.e. the first Card is at the bottom of the Board, in the z-direction)",
        min_length=1,
    )

    sensors: List[Sensor] = pydantic.Field(min_length=1)
    reinforcer_maps: List[ReinforcerMap] = pydantic.Field(default_factory=list)
    effects: List[Effect] = pydantic.Field(default_factory=list)

    @pydantic.model_validator(mode='after')
    def check_node_is_well_formed(self) -> Self:

        # Run basic checks
        # Check Cards have unique IDs
        card_id_to_card = {card.card_id: card for card in self.cards}
        if len(card_id_to_card) != len(self.cards):
            raise ValueError("Cards must have unique IDs.")

        # Check Sensors:
        sensor_id_to_sensor = {sensor.sensor_id: sensor for sensor in self.sensors}
        if len(sensor_id_to_sensor) != len(self.sensors):
            raise ValueError("Sensors must have unique IDs.")
        for sensor in self.sensors:

            # Ensure Sensor references a valid target
            if sensor.card_id and sensor.card_id not in card_id_to_card:
                raise ValueError(f"Sensor {sensor.sensor_id} is not attached to a valid Card or the Board.")

            # Todo: check that the Sensor and Target type are compatible:
            ...

        # Check Reinforcer Maps
        for rmap in self.reinforcer_maps:
            if rmap.sensor_id not in sensor_id_to_sensor:
                raise ValueError(f"ReinforcerMap {rmap.reinforcer_map_id} is not attached to a valid Sensor.")

            # Todo: apply rules for ReinforcerMap-Sensor compatibility
            ...

        return self


# %% NodeGraph
class NodeGraph(pydantic.BaseModel):
    nodes: List[Node]

    # Payment information
    base_payment_usd: PayableMonetaryAmountUsd = pydantic.Field(
        description="The base payment (in USD) that a Participant receives upon successfully completing a TaskRun. Explicitly disclosed to the Participant ahead of time.",
    )
    bonus_rules: List[BonusRule] = pydantic.Field(
        default_factory=list,
        description='A list of bonus rules that apply to the TaskRun. These rules determine additional payments based on Participant behavior during the TaskRun.'
                    'These are not disclosed to the Participant by default, but might be described elsewhere, like the description or in a Node.'
    )

    # Duration
    max_duration_sec: int = pydantic.Field(
        gt=0,
        description="The maximum of time in seconds a Participant has to complete a TaskRun, before it is marked as failed."
    )

    # Metadata that is disclosed to the Participant:
    title: str = pydantic.Field(
        min_length=1,
        description='The title of the Recruitment Platform posting for the task.'
    )
    description: str = pydantic.Field(
        min_length=1,
        description='A detailed description of the task in the Recruitment Platform posting.'
    )
    keywords: List[str] = pydantic.Field(
        description='Keywords that Participants may use to discover this task on the Recruitment Platform.'
    )



# %%
class NodeResult(pydantic.BaseModel):
    """
    Describes the result of a NodePlay.
    """

    node_id: str = pydantic.Field(description='The ID of the Node from which this NodeResult was produced.')
    node_execution_index: int = pydantic.Field(description='The index of the Node execution in the NodeGraph. This is used to identify the order of Node executions in a TaskRun.')

    timestamp_start: DatetimeUTC
    timestamp_end: DatetimeUTC

    action: Action
    runtime_metrics: RuntimeMetrics



from typing import Literal, Annotated, Union, List

import pydantic

from nodekit._internal.models.node_engine.base import NullParameters, DslModel
from nodekit._internal.models.node_engine.cards.base import BaseCard
from nodekit._internal.models.node_engine.fields import TextContent, ColorHexString
from nodekit._internal.models.assets.base import ImageLink


# %% Concrete card classes
class FixationPointCard(BaseCard):
    card_type: Literal['FixationPointCard'] = 'FixationPointCard'
    card_parameters: NullParameters = pydantic.Field(default_factory=NullParameters, frozen=True)


# %%
class MarkdownPagesCard(BaseCard):
    class Parameters(DslModel):
        pages: List[TextContent] = pydantic.Field(
            description='A list of MarkdownContent objects representing the text content on the pages to be displayed.'
        )

    card_type: Literal['MarkdownPagesCard'] = 'MarkdownPagesCard'
    card_parameters: Parameters


# %%
class ImageCard(BaseCard):
    class Parameters(DslModel):
        image_link: ImageLink

    card_type: Literal['ImageCard'] = 'ImageCard'
    card_parameters: Parameters


# %%
class TextCard(BaseCard):
    class Parameters(DslModel):
        content: TextContent
        background_color: ColorHexString = pydantic.Field(
            default='#E6E6E6',
            description='The background color of the TextCard in hexadecimal format.'
        )

    card_type: Literal['TextCard'] = 'TextCard'
    card_parameters: Parameters


# %%
Card = Annotated[
    Union[
        FixationPointCard,
        ImageCard,
        TextCard,
        MarkdownPagesCard,
        # Add other Card types here as needed
    ],
    pydantic.Field(discriminator='card_type')
]



__all__ = [
    "EventTypeEnum",
    "Event",

    # Concrete classes:
    "StartEvent",
    "EndEvent",
    "NodeResultEvent",

    # Client-server models:
    "SubmitEventRequest",
    "SubmitEventResponse",
]

from nodekit._internal.compilers.events import (
    Event,
    StartEvent,
    EndEvent,
    NodeResultEvent,
    SubmitEventRequest,
    SubmitEventResponse,
    EventTypeEnum,
)



from pathlib import Path

import jinja2

from nodekit._internal.models.node_engine.node_graph import NodeGraph
import pydantic


# %%
class CompileHtmlOptions(pydantic.BaseModel):
    event_submission_url: str | None = pydantic.Field(
        description='An endpoint to which Events will be sent. If None, the Events will simply be shown at the end of the Run.',
        default = None,
    )
    start_node_execution_index: int

    run_id: str


def html(
        node_graph: NodeGraph,
        options: CompileHtmlOptions | None = None,
) -> str:
    if options is None:
        options = CompileHtmlOptions(
            event_submission_url=None,
            start_node_execution_index=0,
            run_id='NO_RUN_ID',
        )

    # Render the node sequence using a Jinja2 template
    template_location = Path(__file__).parent / 'node_graph_site_template.j2'
    template = jinja2.Environment(loader=jinja2.FileSystemLoader(template_location.parent)).get_template(template_location.name)

    html_string = template.render(
        dict(
            node_graph=node_graph.model_dump(mode='json'),
            event_submission_url=options.event_submission_url,
            run_id=options.run_id,
            start_node_execution_index=options.start_node_execution_index,
        )
    )

    return html_string


__all__ = [
    "Action",
    "ClickAction",
    "DoneAction",
    "TimeoutAction",
]

from nodekit._internal.models.node_engine.sensors.actions.actions import (
    Action,
    ClickAction,
    DoneAction,
    TimeoutAction,
)



from typing import Literal, Union, Annotated

import pydantic

from nodekit._internal.models.node_engine.base import DslModel, NullValue
from nodekit._internal.models.node_engine.sensors.actions.base import BaseAction


# %%

class ClickAction(BaseAction):
    """
    A fully-qualified description of a Sensor emission.
    """

    class Value(DslModel):
        """
        A fully-qualified description of a Sensor emission.
        """
        click_x: float = pydantic.Field(description='The x-coordinate of the click, in Board units.')
        click_y: float = pydantic.Field(description='The y-coordinate of the click, in Board units.')

    action_type: Literal['ClickAction'] = 'ClickAction'
    action_value: Value


# %%
class DoneAction(BaseAction):
    action_type: Literal['DoneAction'] = 'DoneAction'
    action_value: NullValue = pydantic.Field(default_factory=NullValue, frozen=True)


# %%
class TimeoutAction(BaseAction):
    action_type: Literal['TimeoutAction'] = 'TimeoutAction'
    action_value: NullValue = pydantic.Field(default_factory=NullValue, frozen=True)


# %%
Action = Annotated[
    Union[
        ClickAction,
        DoneAction,
        TimeoutAction,
        # Add other Action types here as needed
    ],
    pydantic.Field(discriminator='action_type')
]



from nodekit._internal.models.node_engine.fields import Timespan
from nodekit._internal.models.node_engine.base import NullParameters
from abc import ABC
import pydantic
from typing import TypeVar, Literal, Any, Annotated, Union

T = TypeVar('T', bound=str)
P = TypeVar('P', bound=pydantic.BaseModel)


class BaseEffect(pydantic.BaseModel, ABC):
    effect_type: T
    effect_parameters: Any
    effect_timespan: Timespan = pydantic.Field(default_factory=lambda: Timespan(start_time_msec=0, end_time_msec=None))


class HidePointerEffect(BaseEffect):
    """
    Effect to hide the pointer during a timespan.
    """
    effect_type: Literal['HidePointerEffect'] = 'HidePointerEffect'
    effect_parameters: NullParameters = pydantic.Field(default_factory=NullParameters)


Effect = Annotated[
    Union[HidePointerEffect],
    pydantic.Field(
        discriminator='effect_type',
    )
]



__all__ = [
    'ConstantBonusRule',
    'compute_bonus',
]

from nodekit._internal.models.node_engine.bonus_policy import (
    ConstantBonusRule,
)

from nodekit._internal.rule_engine.compute_bonus import compute_bonus



import pydantic

from nodekit._internal.models.node_engine.base import DslModel


# %% Board
class Board(DslModel):
    board_width_px: int = pydantic.Field(default=768, gt=0)
    board_height_px: int = pydantic.Field(default=768, gt=0)



from abc import ABC
from typing import Any

import pydantic

from nodekit._internal.models.node_engine.base import DslModel
from nodekit._internal.models.node_engine.fields import BoardRectangle, BoardLocation, Timespan, CardId
from uuid import uuid4


class BaseCard(DslModel, ABC):
    """
    Cards are the atomic elements which constitute a single Node.
    Cards are spatially and temporally bound on the Board.
    Some Cards may have Sensors attached to them, which listen for Participant behavior, and emits an Action when triggered.
    """
    # Identifiers
    card_id: CardId = pydantic.Field(default_factory=uuid4)
    card_type: str
    card_parameters: Any
    # Spatial
    card_shape: BoardRectangle
    card_location: BoardLocation
    # Temporal
    card_timespan: Timespan



import pydantic

from nodekit._internal.models.node_engine.base import DslModel


# %% Measurements
class PixelArea(DslModel):
    width_px: int = pydantic.Field(description="Width of the area in pixels.", ge=0)
    height_px: int = pydantic.Field(description="Height of the area in pixels.", ge=0)


class RuntimeMetrics(DslModel):
    model_config = pydantic.ConfigDict(
        extra='allow',
        frozen=True,
    )

    # Board
    display_area: PixelArea = pydantic.Field(description="Metrics of the display area in which the board is rendered.")
    viewport_area: PixelArea = pydantic.Field(description="Metrics of the viewport in which the board is rendered.")
    board_area: PixelArea = pydantic.Field(description="Metrics of the board area in which the cards are rendered.")

    # User agent string
    user_agent: str = pydantic.Field(description="User agent string of the browser or application rendering the board.")



from abc import ABC
from typing import Dict, Any, Annotated, Union, List, Literal

import pydantic

from nodekit._internal.models.node_engine.fields import SensorId
from nodekit._internal.models.fields import (
    MonetaryAmountUsd,
)


# %%
class BaseBonusRule(pydantic.BaseModel, ABC):
    bonus_rule_type: str
    bonus_rule_parameters: Dict[str, Any]


class ConstantBonusRule(BaseBonusRule):
    """
    A bonus rule that applies a bonus whenever a particular Sensor is triggered, regardless of the Action's value.
    """
    bonus_rule_type: Literal['ConstantBonusRule'] = 'ConstantBonusRule'

    class Parameters(pydantic.BaseModel):
        """
        Parameters for the ConstantBonusRule.
        """
        sensor_id: SensorId = pydantic.Field(
            description='The ID of the sensor to which this bonus rule applies.'
        )
        bonus_amount_usd: MonetaryAmountUsd = pydantic.Field(
            description='The change in bonus amount to apply when the sensor is triggered. Can be positive or negative.'
        )

    bonus_rule_parameters: Parameters


# %%
BonusRule = Annotated[
    Union[ConstantBonusRule],
    pydantic.Field(
        discriminator='bonus_rule_type',
        description='The type of bonus rule to apply.'
    )
]


# %%
class BonusPolicy(pydantic.BaseModel):
    bonus_rules: List[BonusRule] = pydantic.Field(default_factory=list)



from typing import Literal, Union, Annotated

import pydantic

from nodekit._internal.models.node_engine.base import DslModel, NullParameters
from nodekit._internal.models.node_engine.fields import TimeDurationMsec
from nodekit._internal.models.node_engine.sensors.base import BaseSensor

from nodekit._internal.models.node_engine.fields import CardId


# %%
class TimeoutSensor(BaseSensor):
    """
    Attaches to the Board and triggers an Action after a specified timeout period.
    """

    class TimeoutSensorParameters(DslModel):
        timeout_msec: TimeDurationMsec

    sensor_type: Literal['TimeoutSensor'] = 'TimeoutSensor'
    sensor_parameters: TimeoutSensorParameters
    card_id: None = pydantic.Field(default=None, frozen=True)  # Only binds to Board


# %%
class DoneSensor(BaseSensor):
    sensor_type: Literal['DoneSensor'] = 'DoneSensor'
    sensor_parameters: NullParameters = pydantic.Field(default_factory=NullParameters, frozen=True)
    card_id: CardId = pydantic.Field(default=None, frozen=True)  # Only binds to Card


# %%
class ClickSensor(BaseSensor):
    sensor_type: Literal['ClickSensor'] = pydantic.Field(default='ClickSensor', frozen=True)
    sensor_parameters: NullParameters = pydantic.Field(default_factory=NullParameters, frozen=True)
    card_id: CardId = pydantic.Field(default=None, frozen=True)  # Only binds to Card


# %%
Sensor = Annotated[
    Union[
        TimeoutSensor,
        DoneSensor,
        ClickSensor,
        # Add other Sensor types here as needed
    ],
    pydantic.Field(discriminator='sensor_type')
]



from typing import Literal, Annotated, TypeAlias
from uuid import UUID

import pydantic

SpatialSize = Annotated[float, pydantic.Field(
    strict=True, ge=0, le=1, description='A spatial size relative to the smaller extent of the board (width or height, whichever is smaller). For example, a value of 0.5 corresponds to half the smaller extent of the board.'
)]
SpatialPoint = Annotated[float, pydantic.Field(strict=True, ge=-0.5, le=0.5)]
TimeDurationMsec = Annotated[int, pydantic.Field(strict=True, ge=0, description='A duration of time in milliseconds.')]
TimePointMsec = Annotated[int, pydantic.Field(strict=True, ge=0, description='A point in time relative to some start time in milliseconds.')]
CurrencyCode = Literal['USD']
MarkdownString = str

CardId = UUID
SensorId = UUID
NodeId = UUID

# Alpha-supporting hex RGB color string:
ColorHexString = Annotated[
    str,
    pydantic.Field(
        pattern=r'^#(?:[0-9a-fA-F]{3}){1,2}$',
        min_length=7,
        max_length=9,
    )
]


class TextContent(pydantic.BaseModel):
    text: MarkdownString
    text_color: ColorHexString = '#000000'
    font_size: SpatialSize = 0.0175
    justification_horizontal: Literal['left', 'center', 'right'] = 'left'
    justification_vertical: Literal['top', 'center', 'bottom'] = 'top'


class BoardRectangle(pydantic.BaseModel):
    """
    Describes a rectangle on the Board, in Board units.
    """
    width: SpatialSize
    height: SpatialSize


class BoardLocation(pydantic.BaseModel):
    """
    Describes the location of a point on the Board. The coordinates of that location are given under a
     coordinate system where:
    - (0,0) is the center of the Board.
    - A unit of 1 corresponds to the *smaller* extent of the Board (the full width of the Board or the full height of the Board; whichever is smaller.).
    - A positive increase in the x-dimension is rightwards.
    - A positive increase in the y-dimension is upwards.
    """
    x: SpatialPoint
    y: SpatialPoint


class Timespan(pydantic.BaseModel):
    start_time_msec: TimePointMsec = pydantic.Field(
        description="The start time of this Timespan relative to the NodePlay start, in milliseconds."
    )

    end_time_msec: TimePointMsec | None = pydantic.Field(
        description="The time, relative to NodePlay start, when the time span ends, in milliseconds. If None, the time span is open-ended and continues until the end of NodePlay.",
        default=None,
    )

    def check_if_subset_of_other(self, timespan_other: 'Timespan'):
        """
        Check if this Timespan is a subset of another Timespan.
        """
        if self.start_time_msec < timespan_other.start_time_msec:
            return False

        if self.end_time_msec is not None and timespan_other.end_time_msec is not None:
            if self.end_time_msec > timespan_other.end_time_msec:
                return False
        return True

    def is_finite(self) -> bool:
        """
        Check if the Timespan is finite (i.e., has a defined end time).
        """
        return self.end_time_msec is not None
